@@ -55,6 +55,7 @@
 #include "glib-helper.h"
 #include "agent.h"
 #include "storage.h"
+#include "attrib-server.h"
 #include "att.h"
 
 /* Flags Descriptions */
@@ -268,6 +269,10 @@ int adapter_resolve_names(struct btd_adapter *adapter)
 	struct remote_dev_info *dev, match;
 	int err;
 
+	/* Do not attempt to resolve more names if on suspended state */
+	if (adapter->state & STATE_SUSPENDED)
+		return 0;
+
 	memset(&match, 0, sizeof(struct remote_dev_info));
 	bacpy(&match.bdaddr, BDADDR_ANY);
 	match.name_status = NAME_REQUIRED;
@@ -867,6 +872,12 @@ void btd_adapter_class_changed(struct btd_adapter *adapter, uint32_t new_class)
 
 	adapter->dev_class = new_class;
 
+	if (main_opts.attrib_server) {
+		/* Removes service class */
+		class[1] = class[1] & 0x1f;
+		attrib_gap_set(GATT_CHARAC_APPEARANCE, class, 2);
+	}
+
 	emit_property_changed(connection, adapter->path,
 				ADAPTER_INTERFACE, "Class",
 				DBUS_TYPE_UINT32, &new_class);
@@ -881,6 +892,10 @@ void adapter_update_local_name(struct btd_adapter *adapter, const char *name)
 
 	strncpy(dev->name, name, MAX_NAME_LENGTH);
 
+	if (main_opts.attrib_server)
+		attrib_gap_set(GATT_CHARAC_DEVICE_NAME,
+			(const uint8_t *) dev->name, strlen(dev->name));
+
 	if (!adapter->name_stored) {
 		char *name_ptr = dev->name;
 
@@ -947,21 +962,6 @@ struct btd_device *adapter_find_device(struct btd_adapter *adapter,
 	return device;
 }
 
-struct btd_device *adapter_find_connection(struct btd_adapter *adapter,
-						uint16_t handle)
-{
-	GSList *l;
-
-	for (l = adapter->connections; l; l = l->next) {
-		struct btd_device *device = l->data;
-
-		if (device_has_connection(device, handle))
-			return device;
-	}
-
-	return NULL;
-}
-
 static void adapter_update_devices(struct btd_adapter *adapter)
 {
 	char **devices;
@@ -1004,6 +1004,51 @@ static void adapter_emit_uuids_updated(struct btd_adapter *adapter)
 	g_strfreev(uuids);
 }
 
+static uint8_t get_uuid_mask(uuid_t *uuid)
+{
+	if (uuid->type != SDP_UUID16)
+		return 0;
+
+	switch (uuid->value.uuid16) {
+	case DIALUP_NET_SVCLASS_ID:
+	case CIP_SVCLASS_ID:
+		return 0x42;	/* Telephony & Networking */
+	case IRMC_SYNC_SVCLASS_ID:
+	case OBEX_OBJPUSH_SVCLASS_ID:
+	case OBEX_FILETRANS_SVCLASS_ID:
+	case IRMC_SYNC_CMD_SVCLASS_ID:
+	case PBAP_PSE_SVCLASS_ID:
+		return 0x10;	/* Object Transfer */
+	case HEADSET_SVCLASS_ID:
+	case HANDSFREE_SVCLASS_ID:
+		return 0x20;	/* Audio */
+	case CORDLESS_TELEPHONY_SVCLASS_ID:
+	case INTERCOM_SVCLASS_ID:
+	case FAX_SVCLASS_ID:
+	case SAP_SVCLASS_ID:
+	/*
+	 * Setting the telephony bit for the handsfree audio gateway
+	 * role is not required by the HFP specification, but the
+	 * Nokia 616 carkit is just plain broken! It will refuse
+	 * pairing without this bit set.
+	 */
+	case HANDSFREE_AGW_SVCLASS_ID:
+		return 0x40;	/* Telephony */
+	case AUDIO_SOURCE_SVCLASS_ID:
+	case VIDEO_SOURCE_SVCLASS_ID:
+		return 0x08;	/* Capturing */
+	case AUDIO_SINK_SVCLASS_ID:
+	case VIDEO_SINK_SVCLASS_ID:
+		return 0x04;	/* Rendering */
+	case PANU_SVCLASS_ID:
+	case NAP_SVCLASS_ID:
+	case GN_SVCLASS_ID:
+		return 0x02;	/* Networking */
+	default:
+		return 0;
+	}
+}
+
 static int uuid_cmp(const void *a, const void *b)
 {
 	const sdp_record_t *rec = a;
@@ -1025,8 +1070,10 @@ void adapter_service_insert(struct btd_adapter *adapter, void *r)
 	adapter->services = sdp_list_insert_sorted(adapter->services, rec,
 								record_sort);
 
-	if (new_uuid)
-		adapter_ops->add_uuid(adapter->dev_id, &rec->svclass);
+	if (new_uuid) {
+		uint8_t svc_hint = get_uuid_mask(&rec->svclass);
+		adapter_ops->add_uuid(adapter->dev_id, &rec->svclass, svc_hint);
+	}
 
 	adapter_emit_uuids_updated(adapter);
 }
@@ -1802,6 +1849,9 @@ static DBusMessage *find_device(DBusConnection *conn, DBusMessage *msg,
 
 static void agent_removed(struct agent *agent, struct btd_adapter *adapter)
 {
+	adapter_ops->set_io_capability(adapter->dev_id,
+					IO_CAPABILITY_NOINPUTNOOUTPUT);
+
 	adapter->agent = NULL;
 }
 
@@ -1836,6 +1886,8 @@ static DBusMessage *register_agent(DBusConnection *conn, DBusMessage *msg,
 	DBG("Agent registered for hci%d at %s:%s", adapter->dev_id, name,
 			path);
 
+	adapter_ops->set_io_capability(adapter->dev_id, cap);
+
 	return dbus_message_new_method_return(msg);
 }
 
@@ -1928,6 +1980,7 @@ static struct link_key_info *get_key_info(const char *addr, const char *value)
 {
 	struct link_key_info *info;
 	char tmp[3];
+	long int l;
 	int i;
 
 	if (strlen(value) < 36) {
@@ -1950,7 +2003,10 @@ static struct link_key_info *get_key_info(const char *addr, const char *value)
 	info->type = (uint8_t) strtol(tmp, NULL, 10);
 
 	memcpy(tmp, value + 35, 2);
-	info->pin_len = strtol(tmp, NULL, 10);
+	l = strtol(tmp, NULL, 10);
+	if (l < 0)
+		l = 0;
+	info->pin_len = l;
 
 	return info;
 }
@@ -2188,14 +2244,16 @@ static void load_connections(struct btd_adapter *adapter)
 	}
 
 	for (l = conns; l != NULL; l = g_slist_next(l)) {
-		struct hci_conn_info *ci = l->data;
+		bdaddr_t *bdaddr = l->data;
 		struct btd_device *device;
 		char address[18];
 
-		ba2str(&ci->bdaddr, address);
+		ba2str(bdaddr, address);
+		DBG("Adding existing connection to %s", address);
+
 		device = adapter_get_device(connection, adapter, address);
 		if (device)
-			adapter_add_connection(adapter, device, ci->handle);
+			adapter_add_connection(adapter, device);
 	}
 
 	g_slist_foreach(conns, (GFunc) g_free, NULL);
@@ -2462,7 +2520,7 @@ int btd_adapter_stop(struct btd_adapter *adapter)
 
 	while (adapter->connections) {
 		struct btd_device *device = adapter->connections->data;
-		adapter_remove_connection(adapter, device, 0);
+		adapter_remove_connection(adapter, device);
 	}
 
 	if (adapter->scan_mode == (SCAN_PAGE | SCAN_INQUIRY)) {
@@ -2600,6 +2658,10 @@ gboolean adapter_init(struct btd_adapter *adapter)
 		expand_name(adapter->dev.name, MAX_NAME_LENGTH, main_opts.name,
 							adapter->dev_id);
 
+	if (main_opts.attrib_server)
+		attrib_gap_set(GATT_CHARAC_DEVICE_NAME,
+			(const uint8_t *) dev->name, strlen(dev->name));
+
 	sdp_init_services_list(&adapter->bdaddr);
 	load_drivers(adapter);
 	clear_blocked(adapter);
@@ -2985,7 +3047,7 @@ static void dev_prepend_uuid(gpointer data, gpointer user_data)
 void adapter_update_device_from_info(struct btd_adapter *adapter,
 					bdaddr_t bdaddr, int8_t rssi,
 					uint8_t evt_type, const char *name,
-					GSList *services, uint8_t flags)
+					GSList *services, int flags)
 {
 	struct remote_dev_info *dev;
 	gboolean new_dev;
@@ -3006,7 +3068,8 @@ void adapter_update_device_from_info(struct btd_adapter *adapter,
 	g_slist_foreach(services, remove_same_uuid, dev);
 	g_slist_foreach(services, dev_prepend_uuid, dev);
 
-	dev->flags = flags;
+	if (flags >= 0)
+		dev->flags = flags;
 
 	if (name) {
 		g_free(dev->name);
@@ -3140,29 +3203,31 @@ struct agent *adapter_get_agent(struct btd_adapter *adapter)
 }
 
 void adapter_add_connection(struct btd_adapter *adapter,
-				struct btd_device *device, uint16_t handle)
+						struct btd_device *device)
 {
 	if (g_slist_find(adapter->connections, device)) {
-		error("Unable to add connection %d", handle);
+		error("Device is already marked as connected");
 		return;
 	}
 
-	device_add_connection(device, connection, handle);
+	device_add_connection(device, connection);
 
 	adapter->connections = g_slist_append(adapter->connections, device);
 }
 
 void adapter_remove_connection(struct btd_adapter *adapter,
-				struct btd_device *device, uint16_t handle)
+						struct btd_device *device)
 {
 	bdaddr_t bdaddr;
 
+	DBG("");
+
 	if (!g_slist_find(adapter->connections, device)) {
-		error("No matching connection for handle %u", handle);
+		error("No matching connection for device");
 		return;
 	}
 
-	device_remove_connection(device, connection, handle);
+	device_remove_connection(device, connection);
 
 	adapter->connections = g_slist_remove(adapter->connections, device);
 
@@ -3383,11 +3448,10 @@ const char *adapter_any_get_path(void)
 
 const char *btd_adapter_any_request_path(void)
 {
-	if (adapter_any_refcount > 0)
+	if (adapter_any_refcount++ > 0)
 		return adapter_any_path;
 
 	adapter_any_path = g_strdup_printf("%s/any", manager_get_base_path());
-	adapter_any_refcount++;
 
 	return adapter_any_path;
 }
@@ -3543,9 +3607,9 @@ int btd_adapter_read_clock(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 						timeout, clock, accuracy);
 }
 
-int btd_adapter_disconnect_device(struct btd_adapter *adapter, uint16_t handle)
+int btd_adapter_disconnect_device(struct btd_adapter *adapter, bdaddr_t *bdaddr)
 {
-	return adapter_ops->disconnect(adapter->dev_id, handle);
+	return adapter_ops->disconnect(adapter->dev_id, bdaddr);
 }
 
 int btd_adapter_remove_bonding(struct btd_adapter *adapter, bdaddr_t *bdaddr)
@@ -3553,12 +3617,6 @@ int btd_adapter_remove_bonding(struct btd_adapter *adapter, bdaddr_t *bdaddr)
 	return adapter_ops->remove_bonding(adapter->dev_id, bdaddr);
 }
 
-int btd_adapter_request_authentication(struct btd_adapter *adapter,
-							uint16_t handle)
-{
-	return adapter_ops->request_authentication(adapter->dev_id, handle);
-}
-
 int btd_adapter_pincode_reply(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 							const char *pin)
 {
@@ -3577,17 +3635,6 @@ int btd_adapter_passkey_reply(struct btd_adapter *adapter, bdaddr_t *bdaddr,
 	return adapter_ops->passkey_reply(adapter->dev_id, bdaddr, passkey);
 }
 
-int btd_adapter_get_auth_info(struct btd_adapter *adapter, bdaddr_t *bdaddr,
-								uint8_t *auth)
-{
-	return adapter_ops->get_auth_info(adapter->dev_id, bdaddr, auth);
-}
-
-int btd_adapter_read_scan_enable(struct btd_adapter *adapter)
-{
-	return adapter_ops->read_scan_enable(adapter->dev_id);
-}
-
 void btd_adapter_update_local_ext_features(struct btd_adapter *adapter,
 						const uint8_t *features)
 {
@@ -3607,3 +3654,14 @@ int btd_adapter_set_did(struct btd_adapter *adapter, uint16_t vendor,
 {
 	return adapter_ops->set_did(adapter->dev_id, vendor, product, version);
 }
+
+int adapter_create_bonding(struct btd_adapter *adapter, bdaddr_t *bdaddr,
+								uint8_t io_cap)
+{
+	return adapter_ops->create_bonding(adapter->dev_id, bdaddr, io_cap);
+}
+
+int adapter_cancel_bonding(struct btd_adapter *adapter, bdaddr_t *bdaddr)
+{
+	return adapter_ops->cancel_bonding(adapter->dev_id, bdaddr);
+}
