@@ -295,7 +295,7 @@ static void command_complete(uint16_t ogf, uint16_t ocf, int plen, void *data)
 	he = (void *) ptr; ptr += HCI_EVENT_HDR_SIZE;
 
 	he->evt  = EVT_CMD_COMPLETE;
-	he->plen = EVT_CMD_COMPLETE_SIZE + plen; 
+	he->plen = EVT_CMD_COMPLETE_SIZE + plen;
 
 	cc = (void *) ptr; ptr += EVT_CMD_COMPLETE_SIZE;
 
@@ -327,7 +327,7 @@ static void connect_request(struct vhci_conn *conn)
 	he = (void *) ptr; ptr += HCI_EVENT_HDR_SIZE;
 
 	he->evt  = EVT_CONN_REQUEST;
-	he->plen = EVT_CONN_REQUEST_SIZE; 
+	he->plen = EVT_CONN_REQUEST_SIZE;
 
 	cr = (void *) ptr; ptr += EVT_CONN_REQUEST_SIZE;
 
@@ -355,7 +355,7 @@ static void connect_complete(struct vhci_conn *conn)
 	he = (void *) ptr; ptr += HCI_EVENT_HDR_SIZE;
 
 	he->evt  = EVT_CONN_COMPLETE;
-	he->plen = EVT_CONN_COMPLETE_SIZE; 
+	he->plen = EVT_CONN_COMPLETE_SIZE;
 
 	cc = (void *) ptr; ptr += EVT_CONN_COMPLETE_SIZE;
 
@@ -439,7 +439,7 @@ static int scan_enable(uint8_t *data)
 
 	if (!(*data & SCAN_PAGE)) {
 		if (vdev.scan) {
-			g_io_channel_close(vdev.scan);
+			g_io_channel_shutdown(vdev.scan, TRUE, NULL);
 			vdev.scan = NULL;
 		}
 		return 0;
@@ -499,10 +499,13 @@ static void accept_connection(uint8_t *data)
 
 static void close_connection(struct vhci_conn *conn)
 {
-	syslog(LOG_INFO, "Closing connection %s handle %d",
-					batostr(&conn->dest), conn->handle);
+	char addr[18];
 
-	g_io_channel_close(conn->chan);
+	ba2str(&conn->dest, addr);
+	syslog(LOG_INFO, "Closing connection %s handle %d",
+					addr, conn->handle);
+
+	g_io_channel_shutdown(conn->chan, TRUE, NULL);
 	g_io_channel_unref(conn->chan);
 
 	vconn[conn->handle - 1] = NULL;
@@ -968,13 +971,16 @@ static gboolean io_hci_data(GIOChannel *chan, GIOCondition cond, gpointer data)
 {
 	unsigned char buf[HCI_MAX_FRAME_SIZE], *ptr;
 	int type;
-	gsize len;
-	GIOError err;
+	ssize_t len;
+	int fd;
 
 	ptr = buf;
 
-	if ((err = g_io_channel_read(chan, (gchar *) buf, sizeof(buf), &len))) {
-		if (err == G_IO_ERROR_AGAIN)
+	fd = g_io_channel_unix_get_fd(chan);
+
+	len = read(fd, buf, sizeof(buf));
+	if (len < 0) {
+		if (errno == EAGAIN)
 			return TRUE;
 
 		syslog(LOG_ERR, "Read failed: %s (%d)", strerror(errno), errno);
